package explorviz.live_trace_processing.filter.reconstruction;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.concurrent.TimeUnit;

import explorviz.live_trace_processing.Constants;
import explorviz.live_trace_processing.filter.AbstractFilter;
import explorviz.live_trace_processing.filter.IPipeReceiver;
import explorviz.live_trace_processing.filter.reduction.summarization.TracePatternSummarizationFilter;
import explorviz.live_trace_processing.reader.TimeProvider;
import explorviz.live_trace_processing.record.IRecord;
import explorviz.live_trace_processing.record.event.AbstractOperationEventRecord;
import explorviz.live_trace_processing.record.misc.TerminateRecord;
import explorviz.live_trace_processing.record.misc.TimedPeriodRecord;
import explorviz.live_trace_processing.record.trace.Trace;

public final class TraceReconstructionFilter extends AbstractFilter {
	private final long maxTraceTimeout;

	private final Map<Long, TraceReconstructionBuffer> traceId2trace = new ConcurrentSkipListMap<Long, TraceReconstructionBuffer>();

	public TraceReconstructionFilter(final long maxTraceTimeout, final IPipeReceiver sinkReceiver) {
		super(
				new TracePatternSummarizationFilter(TimeUnit.MILLISECONDS.toNanos(990),
						sinkReceiver), Constants.TRACE_RECONSTRUCTION_DISRUPTOR_SIZE,
				Constants.TRACE_RECONSTRUCTION_OUTPUT_BUFFER_SIZE, "Reconstructed traces/sec");
		this.maxTraceTimeout = maxTraceTimeout;
	}

	@Override
	public final void processRecord(final IRecord record) {
		if (record instanceof AbstractOperationEventRecord) {
			final AbstractOperationEventRecord abstractOperationEvent = ((AbstractOperationEventRecord) record);

			final long traceId = abstractOperationEvent.getTraceId();
			final TraceReconstructionBuffer traceBuffer = getBufferForTraceId(abstractOperationEvent
					.getTraceId());
			traceBuffer.insertEvent(abstractOperationEvent);

			if (traceBuffer.isFinished()) {
				traceId2trace.remove(traceId);
				deliver(traceBuffer.toTrace(true));
			}
		} else if (record instanceof Trace) {
			final Trace trace = (Trace) record;
			if (trace.isValid()) {
				deliver(trace);
			} else {
				getBufferForTraceId(trace.getTraceEvents().get(0).getTraceId());
			}
		} else if (record instanceof TimedPeriodRecord) {
			checkForTimeouts(TimeProvider.getCurrentTimestamp());
			periodicFlush(record);
			deliver(record);
		} else if (record instanceof TerminateRecord) {
			terminate();
			deliver(record);
		} else {
			deliver(record);
		}
	}

	private final TraceReconstructionBuffer getBufferForTraceId(final long traceId) {
		TraceReconstructionBuffer traceBuffer = traceId2trace.get(traceId);
		if (traceBuffer == null) {
			traceBuffer = new TraceReconstructionBuffer();
			traceId2trace.put(traceId, traceBuffer);
		}
		return traceBuffer;
	}

	private void checkForTimeouts(final long timestamp) {
		final long traceTimeout = timestamp - maxTraceTimeout;
		final List<Long> traceIdsToRemove = new ArrayList<Long>();

		for (final Entry<Long, TraceReconstructionBuffer> entry : traceId2trace.entrySet()) {
			final TraceReconstructionBuffer traceBuffer = entry.getValue();
			if (traceBuffer.getMaxLoggingTimestamp() <= traceTimeout) {
				deliver(traceBuffer.toTrace(false));
				traceIdsToRemove.add(entry.getKey());
			}
		}

		for (final Long traceIdToRemove : traceIdsToRemove) {
			traceId2trace.remove(traceIdToRemove);
		}
	}

	private void terminate() {
		for (final TraceReconstructionBuffer entry : traceId2trace.values()) {
			deliver(entry.toTrace(false));
		}
		traceId2trace.clear();
	}
}